// Copyright (C) 2019 by Pedro Mendes, Rector and Visitors of the
// University of Virginia, University of Heidelberg, and University
// of Connecticut School of Medicine.
// All rights reserved.

// Copyright (C) 2017 - 2018 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., University of Heidelberg, and University of
// of Connecticut School of Medicine.
// All rights reserved.

// Copyright (C) 2010 - 2016 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., University of Heidelberg, and The University
// of Manchester.
// All rights reserved.

// Copyright (C) 2009 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., EML Research, gGmbH, University of Heidelberg,
// and The University of Manchester.
// All rights reserved.

#include "CQFunctionsWidget.h"

#include <QHeaderView>
#include <QClipboard>
#include <QKeyEvent>

#include "copasi.h"

#include "qtUtilities.h"
#include "CQMessageBox.h"

#include "model/CModel.h"
#include "CopasiDataModel/CDataModel.h"
#include "copasi/core/CRootContainer.h"
#include "function/CFunctionDB.h"

/*
 *  Constructs a CQFunctionsWidget which is a child of 'parent', with the
 *  name 'name'.'
 */
CQFunctionsWidget::CQFunctionsWidget(QWidget *parent, const char *name)
  : CopasiWidget(parent, name)
{
  setupUi(this);
  //Create Source Data Model.
  mpFunctionDM = new CQFunctionDM(this, mpDataModel);
  //Create the Proxy Model for sorting/filtering and set its properties.
  mpProxyModel = new CQSortFilterProxyModel();
  mpProxyModel->setSortCaseSensitivity(Qt::CaseInsensitive);
  mpProxyModel->setFilterKeyColumn(-1);
#if QT_VERSION >= 0x050000
  mpTblFunctions->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
  mpTblFunctions->verticalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
  mpTblFunctions->verticalHeader()->hide();
  mpTblFunctions->sortByColumn(COL_ROW_NUMBER, Qt::AscendingOrder);
  // Connect the table widget
  connect(mpFunctionDM, SIGNAL(notifyGUI(ListViews::ObjectType, ListViews::Action, const CCommonName &)),
          this, SLOT(protectedNotify(ListViews::ObjectType, ListViews::Action, const CCommonName &)));
  connect(mpFunctionDM, SIGNAL(dataChanged(const QModelIndex &, const QModelIndex &)),
          this, SLOT(dataChanged(const QModelIndex &, const QModelIndex &)));
  connect(mpLEFilter, SIGNAL(textChanged(const QString &)),
          this, SLOT(slotFilterChanged()));
}

/*
 *  Destroys the object and frees any allocated resources
 */
CQFunctionsWidget::~CQFunctionsWidget()
{
  pdelete(mpProxyModel);
  pdelete(mpFunctionDM);
  // no need to delete child widgets, Qt does it all for us
}

void CQFunctionsWidget::slotBtnNewClicked()
{
  mpFunctionDM->insertRow(mpFunctionDM->rowCount(), QModelIndex());
  updateDeleteBtns();
}

void CQFunctionsWidget::slotBtnDeleteClicked()
{
  if (mpTblFunctions->hasFocus())
    {deleteSelectedFunctions();}

  updateDeleteBtns();
}

void CQFunctionsWidget::deleteSelectedFunctions()
{
  const QItemSelectionModel *pSelectionModel = mpTblFunctions->selectionModel();
  QModelIndexList mappedSelRows;
  size_t i, imax = mpFunctionDM->rowCount();

  for (i = 0; i < imax; i++)
    {
      if (pSelectionModel->isRowSelected((int) i, QModelIndex()))
        {
          mappedSelRows.append(mpProxyModel->mapToSource(mpProxyModel->index((int) i, 0)));
        }
    }

  if (mappedSelRows.empty())
    {return;}

  mpFunctionDM->removeRows(mappedSelRows);
  updateDeleteBtns();
}

void CQFunctionsWidget::slotBtnClearClicked()
{
  int ret = CQMessageBox::question(this, tr("Confirm Delete"), "Delete all non-built-in, non-mass-acion, Functions?",
                                   QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

  if (ret == QMessageBox::Yes)
    {
      QModelIndexList mappedSelRows;
      size_t i, imax = mpFunctionDM->rowCount();

      for (i = 0; i < imax; i++)
        {
          mappedSelRows.append(mpFunctionDM->index((int) i, 0));
        }

      if (mappedSelRows.empty())
        return;

      mpFunctionDM->removeRows(mappedSelRows);
    }

  updateDeleteBtns();
}

bool CQFunctionsWidget::updateProtected(ListViews::ObjectType objectType, ListViews::Action action, const CCommonName & cn)
{
  if (!mIgnoreUpdates && isVisible())
    {
      enterProtected();
    }

  return true;
}

bool CQFunctionsWidget::leaveProtected()
{
  return true;
}

bool CQFunctionsWidget::enterProtected()
{
  if (mpTblFunctions->selectionModel() != NULL)
    {
      disconnect(mpTblFunctions->selectionModel(), SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
                 this, SLOT(slotSelectionChanged(const QItemSelection &, const QItemSelection &)));
    }

  mpFunctionDM->setDataModel(mpDataModel);
  mpProxyModel->setSourceModel(mpFunctionDM);
  //Set Model for the TableView
  mpTblFunctions->setModel(NULL);
  mpTblFunctions->setModel(mpProxyModel);
  connect(mpTblFunctions->selectionModel(), SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
          this, SLOT(slotSelectionChanged(const QItemSelection &, const QItemSelection &)));
  updateDeleteBtns();
  mpTblFunctions->resizeColumnsToContents();
  return true;
}

void CQFunctionsWidget::updateDeleteBtns()
{
  bool selected = false;
  QModelIndexList selRows = mpTblFunctions->selectionModel()->selectedRows();

  if (selRows.size() == 0)
    selected = false;
  else
    {
      if (selRows.size() == 1)
        {
          if (mpFunctionDM->isDefaultRow(mpProxyModel->mapToSource(selRows[0])))
            selected = false;
          else
            selected = true;
        }
      else
        selected = true;
    }

  mpBtnDelete->setEnabled(selected);

  if (mpProxyModel->rowCount() - 1)
    mpBtnClear->setEnabled(true);
  else
    mpBtnClear->setEnabled(false);
}

void CQFunctionsWidget::slotSelectionChanged(const QItemSelection &C_UNUSED(selected),
    const QItemSelection &C_UNUSED(deselected))
{
  updateDeleteBtns();
}

void CQFunctionsWidget::dataChanged(const QModelIndex &C_UNUSED(topLeft),
                                    const QModelIndex &C_UNUSED(bottomRight))
{
  mpTblFunctions->resizeColumnsToContents();
  updateDeleteBtns();
}

void CQFunctionsWidget::slotDoubleClicked(const QModelIndex proxyIndex)
{
  QModelIndex index = mpProxyModel->mapToSource(proxyIndex);

  if (index.row() < 0)
    return;

  if (mpFunctionDM->isDefaultRow(index))
    {
      slotBtnNewClicked();
    }

  CDataVector < CFunction > * pVector = dynamic_cast< CDataVector < CFunction > * >(mpObject);

  if (pVector != NULL &&
      index.row() < pVector->size())
    {
      mpListView->switchToOtherWidget(ListViews::WidgetType::FunctionDetail, pVector->operator [](index.row()).getCN());
    }
}

void CQFunctionsWidget::keyPressEvent(QKeyEvent *ev)
{
  if (ev->key() == Qt::Key_Delete)
    slotBtnDeleteClicked();
  else if (ev->key() == Qt::Key_C && (ev->modifiers() & Qt::ControlModifier))
    {
      QModelIndexList selRows = mpTblFunctions->selectionModel()->selectedRows(0);

      if (selRows.empty())
        {return;}

      QString str;
      QModelIndexList::const_iterator i;

      for (i = selRows.begin(); i != selRows.end(); ++i)
        {
          for (int x = 0; x < mpFunctionDM->columnCount(); ++x)
            {
              if (!mpTblFunctions->isColumnHidden(x))
                {
                  if (!str.isEmpty())
                    str += "\t";

                  str += mpFunctionDM->index(mpProxyModel->mapToSource(*i).row(), x).data().toString();
                }
            }

          str += "\n";
        }

      QApplication::clipboard()->setText(str);
    }
}

void CQFunctionsWidget::slotFilterChanged()
{
  QRegExp regExp(mpLEFilter->text() + "|New Function", Qt::CaseInsensitive, QRegExp::RegExp);
  mpProxyModel->setFilterRegExp(regExp);
}
